// DO NOT EDIT! This test has been generated by /html/canvas/tools/gentest.py.
// OffscreenCanvas test in a worker:2d.pattern.basic.zerocanvas
// Description:
// Note:

importScripts("/resources/testharness.js");
importScripts("/html/canvas/resources/canvas-tests.js");

var t = async_test("");
var t_pass = t.done.bind(t);
var t_fail = t.step_func(function(reason) {
    throw reason;
});
t.step(function() {

  var canvas = new OffscreenCanvas(100, 50);
  var ctx = canvas.getContext('2d');

  canvas.width = 0;
  canvas.height = 10;
  _assertSame(canvas.width, 0, "canvas.width", "0");
  _assertSame(canvas.height, 10, "canvas.height", "10");
  assert_throws_dom("INVALID_STATE_ERR", function() { ctx.createPattern(canvas, 'repeat'); });

  canvas.width = 10;
  canvas.height = 0;
  _assertSame(canvas.width, 10, "canvas.width", "10");
  _assertSame(canvas.height, 0, "canvas.height", "0");
  assert_throws_dom("INVALID_STATE_ERR", function() { ctx.createPattern(canvas, 'repeat'); });

  canvas.width = 0;
  canvas.height = 0;
  _assertSame(canvas.width, 0, "canvas.width", "0");
  _assertSame(canvas.height, 0, "canvas.height", "0");
  assert_throws_dom("INVALID_STATE_ERR", function() { ctx.createPattern(canvas, 'repeat'); });
  t.done();
});
done();
