// AUTO-GENERATED - DO NOT EDIT. See src/common/tools/gen_listings_and_webworkers.ts.

import { g } from '../../../../../../api/validation/encoding/cmds/render/setPipeline.spec.js';
import { wrapTestGroupForWorker } from '../../../../../../../common/runtime/helper/wrap_for_worker.js';

wrapTestGroupForWorker(g);
